/* -*- Mode: C; tab-width: 4; indent-tabs-mode: t; c-basic-offset: 4 -*- */
/* NetworkManager system settings service
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2008 - 2019 Mikhail Efremov <sem@altlinux.org>
 */

#include "nms-etcnet-alt-common.h"
#include "nms-etcnet-alt-connection.h"
#include "nms-etcnet-alt-reader.h"

G_DEFINE_TYPE (NMEtcnetConnection, nm_etcnet_connection, NM_TYPE_SETTINGS_CONNECTION)

#define NM_ETCNET_CONNECTION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), NM_TYPE_ETCNET_CONNECTION, NMEtcnetConnectionPrivate))

typedef struct
{
	gboolean can_autoconnect;
} NMEtcnetConnectionPrivate;

NMEtcnetConnection *
nm_etcnet_connection_new (const char *iface_path, const char *profile, NMDeviceType dev_type,
                          GHashTable *wsconf, gboolean *ignored)
{
	NMConnection *connection;
	NMEtcnetConnection *exported = NULL;
	gboolean can_autoconnect = FALSE;
	NMEtcnetConnectionPrivate *priv;
	GError *error = NULL;

	g_return_val_if_fail (iface_path != NULL, NULL);
	g_return_val_if_fail (ignored != NULL, NULL);

	connection = read_connection (iface_path, profile, dev_type, wsconf,
									ignored, &can_autoconnect, &error);

	if (!connection) {
		if (!*ignored && error)
			PLUGIN_ERROR("read connection", error);
		goto done;
	}

	exported = (NMEtcnetConnection *) g_object_new (NM_TYPE_ETCNET_CONNECTION, NULL);
	if (!exported) {
		goto done;
	}
	if (!nm_settings_connection_update (NM_SETTINGS_CONNECTION (exported),
	                                    connection,
	                                    NM_SETTINGS_CONNECTION_PERSIST_MODE_KEEP,
	                                    NM_SETTINGS_CONNECTION_COMMIT_REASON_NONE,
	                                    "etcnet-update",
	                                    &error)) {
		PLUGIN_ERROR("update connection", error);
		g_object_unref (exported);
		exported = NULL;
		goto done;
	}

done:
	if (connection)
		g_object_unref (connection);

	if (error)
		g_error_free(error);

	if (exported) {
		priv = NM_ETCNET_CONNECTION_GET_PRIVATE (exported);
		priv->can_autoconnect = can_autoconnect;
	}

	return exported;
}

gboolean
nm_etcnet_connection_get_can_autoconnect (NMEtcnetConnection *self)
{
	g_return_val_if_fail (NM_IS_ETCNET_CONNECTION (self), FALSE);

	return NM_ETCNET_CONNECTION_GET_PRIVATE (self)->can_autoconnect;
}

const char *
nm_etcnet_connection_get_id (NMEtcnetConnection *self)
{
	NMSettingConnection *s_con;

	g_return_val_if_fail (NM_IS_ETCNET_CONNECTION (self), NULL);

	s_con = nm_connection_get_setting_connection (
			  nm_settings_connection_get_connection (NM_SETTINGS_CONNECTION (self)));
	g_return_val_if_fail (s_con != NULL, NULL);
	return nm_setting_connection_get_id (s_con);
}

/* GObject */

static void
nm_etcnet_connection_init (NMEtcnetConnection *connection)
{
}

static void
finalize (GObject *object)
{
	nm_connection_clear_secrets (NM_CONNECTION (object));

	G_OBJECT_CLASS (nm_etcnet_connection_parent_class)->finalize (object);
}

static void
nm_etcnet_connection_class_init (NMEtcnetConnectionClass *etcnet_connection_class)
{
	GObjectClass *object_class = G_OBJECT_CLASS (etcnet_connection_class);

	g_type_class_add_private (etcnet_connection_class, sizeof (NMEtcnetConnectionPrivate));

	/* Virtual methods */
	object_class->finalize = finalize;
}
