/* NetworkManager system settings service
 *
 * Based on ifcfg-fedora and ifcfg-suse code
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * (C) Copyright 2008 - 2019 Mikhail Efremov <sem@altlinux.org>
 */

#include "nms-etcnet-alt-common.h"
#include <sys/stat.h>

#include <libudev.h>

#include "nms-etcnet-alt-plugin.h"
#include "nms-etcnet-alt-reader.h"
#include "nms-etcnet-alt-shvar.h"
#include "nms-etcnet-alt-connection.h"
#include "nm-config.h"
#include "platform/nm-platform.h"
#include "nm-udev-aux/nm-udev-utils.h"

#include <arpa/inet.h>

#define DEFAULT_PROFILE_KEY "#"

#define ETCNET_PATH SYSCONFDIR"/net"
#define ETCNET_IFACES_PATH ETCNET_PATH"/ifaces/"

typedef struct
{
	GFileMonitor *monitor;
	guint id;
}FileMonitor;

typedef struct
{
	SettingsPluginEtcnet *plugin;

	char *iface_path;
	gboolean unmanaged;
	NMDeviceType dev_type;
	GHashTable *connections;

	FileMonitor iface_monitor;
}NMEtcnetDevice;

typedef struct
{
	NMUdevClient *udev_client;

	GHashTable *devices;
	char *ifaces_path;
	gboolean need_monitor_files;
	FileMonitor ifaces_monitor;

	char *current_profile;
} SettingsPluginEtcnetPrivate;

struct _SettingsPluginEtcnet {
	NMSettingsPlugin parent;
	SettingsPluginEtcnetPrivate _priv;
};

struct _SettingsPluginEtcnetClass {
	NMSettingsPluginClass parent;
};

G_DEFINE_TYPE (SettingsPluginEtcnet, settings_plugin_etcnet, NM_TYPE_SETTINGS_PLUGIN)

#define SETTINGS_PLUGIN_ETCNET_GET_PRIVATE(self) _NM_GET_PRIVATE (self, SettingsPluginEtcnet, SETTINGS_IS_PLUGIN_ETCNET)

static SettingsPluginEtcnet *settings_plugin_etcnet_get (void);
NM_DEFINE_SINGLETON_GETTER (SettingsPluginEtcnet, settings_plugin_etcnet_get, SETTINGS_TYPE_PLUGIN_ETCNET);

static void update_device_config(NMEtcnetDevice *etcnet_device, guint *out_unmanaged_changed);

static char *
get_iface_path(const char *ifaces_path, const char *iface)
{
	return g_build_filename (ifaces_path, iface, NULL);
}

static const char *
iface_config_files[]=
{
	"options",
	"ipv4address",
	"ipv4route",
	"ipv6address",
	"ipv6route",
	"iplink",
	"wpa_supplicant.conf",
	NULL
};

static const char *
ignore_suffixes[]=
{
	"~",
	".bak",
	".orig",
	".rej",
	".rpmnew",
	".rpmsave",
	".rpmorig",
	NULL
};

static void
start_monitor (FileMonitor *file_monitor, const char *filepath, GCallback cb_func, gpointer user_data)
{
	GFile *file;
	GFileMonitor *monitor;

	g_return_if_fail(file_monitor != NULL);
	g_return_if_fail(filepath != NULL);

	file = g_file_new_for_path (filepath);
	monitor = g_file_monitor (file, G_FILE_MONITOR_NONE, NULL, NULL);
	g_object_unref (file);

	if (monitor) {
		file_monitor->id = g_signal_connect (monitor, "changed", cb_func, user_data);
		file_monitor->monitor = monitor;
	}
}

static void
stop_monitor (FileMonitor *file_monitor)
{
	g_return_if_fail(file_monitor != NULL);

	if (file_monitor->monitor) {
		if (file_monitor->id)
			g_signal_handler_disconnect (file_monitor->monitor, file_monitor->id);

		g_file_monitor_cancel (file_monitor->monitor);
		g_object_unref (file_monitor->monitor);

		file_monitor->id = 0;
		file_monitor->monitor = NULL;
	}
}

static const char *
etcnet_dev_get_iface_name (const NMEtcnetDevice* etcnet_device)
{
	const char *p = NULL;
	g_return_val_if_fail (etcnet_device != NULL, NULL);

	if (etcnet_device->iface_path) {
		p = strrchr (etcnet_device->iface_path, '/');
		if (p)
			++p;
	}

	return p;
}

#define NETPROFILE_OPTION "netprofile"

/* TODO: Call selectprofile script for each iface */
static char *
get_current_profile (void)
{
	char *line = NULL;
	char *profile = NULL;
	gsize line_len = 0;
	GError *error = NULL;

	profile = getenv ("NETPROFILE");
	if (profile)
		return g_strdup (profile);

	if (g_file_get_contents ("/proc/cmdline", &line, &line_len, &error)) {
		char *b, *e;

		if (line[line_len - 1] == '\n')
			line[line_len - 1] = '\0';
		b = strstr (line, NETPROFILE_OPTION"=");
		if (b) {
			b += strlen (NETPROFILE_OPTION"=");
			e = strpbrk (b, " \t");
			if(e)
				*e = '\0';
			profile = g_strdup (b);
		}
	} else
		PLUGIN_ERROR ("Couldn't read /proc/cmdline", error);

	if (error)
		g_error_free(error);

	g_free (line);

	return profile ?: g_strdup ("#");
}

static void
get_unmanaged_devices_cb (gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) val;
	GSList **list = (GSList **) user_data;

	if(etcnet_device->unmanaged) {
		char *spec = NULL;

		spec = g_strdup_printf ("interface-name:%s", etcnet_dev_get_iface_name (etcnet_device));

		if (spec)
			*list = g_slist_prepend (*list, spec);
	}
}

static GSList *
get_unmanaged_specs (NMSettingsPlugin *config)
{
	GSList *list = NULL;
	SettingsPluginEtcnet *self = SETTINGS_PLUGIN_ETCNET (config);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (self);


	g_hash_table_foreach (priv->devices,
				 get_unmanaged_devices_cb, &list);
	return list;
}

static gboolean
is_file_ignored(const char *filename)
{
	const char **fname, **suffix;
	size_t fname_len;

	g_return_val_if_fail (filename != NULL, TRUE);

	if(filename[0] == '#' || filename[0] == '.')
		return TRUE;

	for (suffix = ignore_suffixes; *suffix != NULL; ++suffix) {
		if (g_str_has_suffix (filename, *suffix))
			return TRUE;
	}

	fname_len = strlen (filename);
	for(fname = iface_config_files; *fname != NULL; ++fname) {
		size_t len = strlen(*fname);
		if(len <= fname_len &&
			    !strncmp(*fname, filename, len) &&
			    /* Exactly match or profiled name */
			    (filename[len] == '\0' || filename[len] == '#'))
			return FALSE;
	}

	return TRUE;
}

static char *
get_profile(const char *filename)
{
	char *p;

	g_return_val_if_fail (filename != NULL, NULL);

	p = strchr(filename, '#');
	if(p != NULL)
		return g_strdup(p + 1);

	return NULL;
}

static char *
find_profile(GDir *iface_dir)
{
	const char *file_name;
	g_return_val_if_fail (iface_dir != NULL, NULL);

	while((file_name = g_dir_read_name(iface_dir)))	{
		char *profile;
		if(is_file_ignored(file_name)) {
			continue;
		}
		profile = get_profile(file_name);
		if(profile != NULL)
			return profile;
	}

	return NULL;
}

static gint
compare_str(gconstpointer a, gconstpointer b)
{
	return strcmp(a, b);
}

static GSList *
get_all_profiles(const char *iface_path)
{
	GDir *iface_dir;
	GSList *profiles_list = NULL;
	char *profile = NULL;

	iface_dir = g_dir_open(iface_path, 0, NULL);
	if (!iface_dir)	{
		return NULL;
	}

	while((profile = find_profile(iface_dir))) {
		if(!g_slist_find_custom(profiles_list, profile, compare_str)) {
			profiles_list = g_slist_prepend(profiles_list, profile);
		} else {
			g_free(profile);
		}
	}

	return profiles_list;
}

static gboolean
is_profile_equal(gconstpointer a, gconstpointer b)
{
	const char *pfl1 = (const char *)a;
	const char *pfl2 = (const char *)b;

	if(!pfl1 && !pfl2)
		return TRUE;
	if(pfl1 && pfl2)
		return strcmp(pfl1, pfl2) ? FALSE : TRUE;
	return FALSE;
}

static void
device_dir_changed (GFileMonitor *monitor,
		   GFile *file,
		   GFile *other_file,
		   GFileMonitorEvent event_type,
		   gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) user_data;
	char *name;

	name = g_file_get_basename (file);
	if ( !is_file_ignored (name)) {
		switch (event_type) {
			case G_FILE_MONITOR_EVENT_DELETED:
			case G_FILE_MONITOR_EVENT_CHANGES_DONE_HINT:
				update_device_config(etcnet_device, NULL);
				break;
			default:
				break;
		}
	}
	g_free (name);
}

static gboolean
is_device_path_equal (gpointer key, gpointer value, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) value;
	char *path = (char *)user_data;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);
	g_return_val_if_fail (path != NULL, FALSE);

	return strcmp (etcnet_device->iface_path, path) ? FALSE : TRUE;
}

static void
ifaces_dir_changed_cb (GFileMonitor *monitor,
		   GFile *file,
		   GFile *other_file,
		   GFileMonitorEvent event_type,
		   gpointer user_data)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (user_data);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);
	char *path;
	NMEtcnetDevice *etcnet_device = NULL;

	path = g_file_get_path(file);
	if(path)
		etcnet_device = g_hash_table_find (priv->devices, is_device_path_equal, path);

	switch (event_type) {
	case G_FILE_MONITOR_EVENT_DELETED:
		if(etcnet_device) {
			/* Check for need_monitor_files not needed:
			 * ifaces_dir_changed_cb() itself is a callback for a file monitor and
			 * if it is called then need_monitor_files is TRUE always */
			stop_monitor(&etcnet_device->iface_monitor);
		}
		break;
	case G_FILE_MONITOR_EVENT_CREATED:
		if(etcnet_device) {
			/* See comment for stop_monitor() above */
			start_monitor(&etcnet_device->iface_monitor, etcnet_device->iface_path,
				G_CALLBACK (device_dir_changed), etcnet_device);
			update_device_config(etcnet_device, NULL);
		}
		break;
	default:
		break;
	}

	g_free(path);
}

static gboolean
add_one_connection_for_device(NMEtcnetDevice *etcnet_device, const char *profile, GHashTable *wsconf)
{
	NMEtcnetConnection *connection;
	gboolean ignored;
	gboolean readed = FALSE;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);
	g_return_val_if_fail (etcnet_device->connections != NULL, FALSE);

	connection = nm_etcnet_connection_new (etcnet_device->iface_path, profile, etcnet_device->dev_type,
                                           wsconf, &ignored);
	if(ignored || connection)
		readed = TRUE;
	if(connection) {
		const char *ssid = NULL;
		char *connection_key = NULL;

		if (ignored) {
			PLUGIN_WARNING ("Ignored, but readed connection");
			g_object_unref (connection);
			/* Connection was readed */
			return TRUE;
		}
		if(profile == NULL)
			profile = DEFAULT_PROFILE_KEY;
		if(wsconf)
			ssid = (const char *)g_hash_table_lookup(wsconf, "ssid");
		connection_key = ssid ? g_strdup_printf ("%s(%s)", profile, ssid) : g_strdup (profile);
		g_hash_table_insert (etcnet_device->connections, connection_key, connection);
		PLUGIN_INFO ("%s: Readed connection \"%s\"", etcnet_dev_get_iface_name (etcnet_device),
				nm_etcnet_connection_get_id (connection));
	}

	return readed;
}

static gboolean
add_connections_for_device (NMEtcnetDevice *etcnet_device, const char *profile)
{
	GHashTable *wsconf = NULL;
	GSList *wsconfig_list = NULL;
	GSList *list_item;
	guint num_readed = 0;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);

	list_item = wsconfig_list = read_supplicant_config(etcnet_device->iface_path, profile);
	do {
		if(list_item)
			wsconf = (GHashTable *)list_item->data;
		if (add_one_connection_for_device (etcnet_device, profile, wsconf))
			++num_readed;
		list_item = g_slist_next(list_item);
	} while(list_item);
	clear_supplicant_config(wsconfig_list);

	return num_readed > 0;
}

static GSList *
get_connections_for_profile (NMEtcnetDevice *dev, const char *profile, gboolean get_default)
{
	GSList *con_list = NULL;
	GList *keys_list, *iter;
	NMEtcnetConnection *connection;
	size_t len;

	g_return_val_if_fail (dev != NULL, NULL);

	if (!profile)
		profile = DEFAULT_PROFILE_KEY;
	switch (dev->dev_type) {
		case NM_DEVICE_TYPE_ETHERNET:
		case NM_DEVICE_TYPE_BRIDGE:
			connection = (NMEtcnetConnection *) g_hash_table_lookup (dev->connections,
					profile);
			if (connection)
				con_list = g_slist_prepend (con_list, connection);
			break;

		case NM_DEVICE_TYPE_WIFI:
			keys_list = g_hash_table_get_keys (dev->connections);
			if (!keys_list)
				break;
			len = strlen (profile);
			for (iter = keys_list; iter; iter = g_list_next (iter)) {
				const char *key = (const char *)iter->data;

				if (strncmp (profile, key, len) == 0) {
					connection = (NMEtcnetConnection *) g_hash_table_lookup (dev->connections,
							key);
					if (connection)
						con_list = g_slist_prepend (con_list, connection);
				}
			}
			break;

		default:
			PLUGIN_WARNING ("Unknown device type: %d", dev->dev_type);
			break;
	}

	/* If no not-ignored connections for current profile and no was readed
	 * (no ignored connections for profile) then use connections for default
	 * profile.
	 */
	if (!con_list && get_default && strcmp (profile, DEFAULT_PROFILE_KEY))
		con_list = get_connections_for_profile (dev, NULL, FALSE);

	return con_list;
}

static gboolean
read_connections_for_one_iface(NMEtcnetDevice *etcnet_device)
{
	gboolean default_readed = FALSE;
	gboolean curr_profiled_readed = FALSE;
	GSList *con_list;
	const char *curr_profile;

	g_return_val_if_fail (etcnet_device != NULL, FALSE);

	curr_profile = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (etcnet_device->plugin)->current_profile;

	if(!g_file_test(etcnet_device->iface_path, G_FILE_TEST_IS_DIR))	{
		default_readed = add_one_connection_for_device (etcnet_device, NULL, NULL);
	} else {
		GSList *profiles, *iter;

		default_readed = add_connections_for_device (etcnet_device, NULL);
		profiles = get_all_profiles(etcnet_device->iface_path);
		for (iter = profiles; iter; iter = g_slist_next (iter)) {
			char *profile_name = (char *)iter->data;

			if (add_connections_for_device (etcnet_device, profile_name) &&
				    curr_profile && !strcmp (curr_profile, profile_name))
				curr_profiled_readed = TRUE;

			g_free(profile_name);
		}

		if(profiles)
			g_slist_free(profiles);
	}

	/* Find connections for current profile */
	con_list = get_connections_for_profile (etcnet_device,
			curr_profile,
			!curr_profiled_readed);
	if (con_list) {
		GSList *iter;
		NMEtcnetConnection *connection;

		etcnet_device->unmanaged = FALSE;
		for (iter = con_list; iter; iter = g_slist_next (iter)) {
			connection = (NMEtcnetConnection *)iter->data;
			if (nm_etcnet_connection_get_can_autoconnect (connection)) {
				NMSettingConnection *s_con;

				s_con = nm_connection_get_setting_connection (
					      nm_settings_connection_get_connection (NM_SETTINGS_CONNECTION (connection)));
				if (s_con)
					g_object_set (s_con, NM_SETTING_CONNECTION_AUTOCONNECT, TRUE, NULL);
				else
					PLUGIN_WARNING ("Can't setup autoconnect property");
			}
		}
	} else
		/* Either default profile or some other current profile was readed,
		 * but connections list is empty means that device is unmanaged
		 */
		etcnet_device->unmanaged = default_readed || curr_profiled_readed;

	return TRUE;
}

static void
connection_added_cb(gpointer key, gpointer val, gpointer user_data)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (user_data);
	NMEtcnetConnection *connection = (NMEtcnetConnection *) val;
	_nm_settings_plugin_emit_signal_connection_added (NM_SETTINGS_PLUGIN (plugin), NM_SETTINGS_CONNECTION (connection));
}

static void
connection_removed_cb(gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetConnection *connection = (NMEtcnetConnection *)val;
	nm_settings_connection_signal_remove (NM_SETTINGS_CONNECTION (connection));
}

static NMEtcnetDevice *
nm_etcnet_device_new (SettingsPluginEtcnet *plugin, const char *ifaces_path, const char *iface, NMDeviceType dev_type)
{
	NMEtcnetDevice *etcnet_device;
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);

	g_return_val_if_fail (ifaces_path != NULL, NULL);
	g_return_val_if_fail (iface != NULL, NULL);

	if (dev_type != NM_DEVICE_TYPE_ETHERNET &&
		dev_type != NM_DEVICE_TYPE_WIFI &&
		dev_type != NM_DEVICE_TYPE_BRIDGE)
		return NULL;

	etcnet_device = (NMEtcnetDevice *)g_malloc0(sizeof(NMEtcnetDevice));
	if(!etcnet_device) {
		PLUGIN_WARNING ("not enough memory for NMEtcnetDevice struct.");
		return NULL;
	}

	etcnet_device->plugin = plugin;
	etcnet_device->iface_path = get_iface_path(ifaces_path, iface);
	etcnet_device->unmanaged = FALSE;
	etcnet_device->dev_type = dev_type;
	etcnet_device->connections = NULL;

	/* We use g_malloc0(), so iface_monitor is zero'ed even if
	 * need_monitor_files is FALSE */
	if (priv->need_monitor_files) {
		start_monitor(&etcnet_device->iface_monitor, etcnet_device->iface_path,
					G_CALLBACK (device_dir_changed), etcnet_device);
	}

	return etcnet_device;
}

static void
nm_etcnet_device_free (gpointer data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *)data;
	g_return_if_fail (etcnet_device != NULL);

	g_free(etcnet_device->iface_path);

	if (etcnet_device->connections) {
		g_hash_table_foreach (etcnet_device->connections, connection_removed_cb, NULL);
		g_hash_table_destroy (etcnet_device->connections);
	}

	stop_monitor(&etcnet_device->iface_monitor);

	g_free(etcnet_device);
}

static void
update_device_config (NMEtcnetDevice *etcnet_device, guint *out_unmanaged_changed)
{
	gboolean old_unmanaged;

	g_return_if_fail (etcnet_device != NULL);

	old_unmanaged = etcnet_device->unmanaged;

	/* Reset all connections
	 * May be we must update connections, not reset them
	 * But is this really worth the efforts?
	 */
	if (etcnet_device->connections) {
		g_hash_table_foreach (etcnet_device->connections, connection_removed_cb, NULL);
		g_hash_table_destroy (etcnet_device->connections);
	}

	etcnet_device->connections = g_hash_table_new_full(g_str_hash, is_profile_equal, g_free, NULL);

	if (!read_connections_for_one_iface (etcnet_device))
		return;

	if (!etcnet_device->unmanaged) {
		if (g_hash_table_size (etcnet_device->connections))
			g_hash_table_foreach (etcnet_device->connections, connection_added_cb, etcnet_device->plugin);
		else {
			PLUGIN_INFO ("No connections for device %s found",
						 etcnet_dev_get_iface_name (etcnet_device));
		}
	} else {
		PLUGIN_INFO ("Device %s is unmanaged", etcnet_dev_get_iface_name (etcnet_device));
	}

	if (etcnet_device->unmanaged != old_unmanaged) {
		/* If out_unmanaged_changed is not NULL then
		 * don't emit the signal.
		 * The signal will be emitted later if at least
		 * one device changed its 'unmanaged' status.
		 */
		if (out_unmanaged_changed)
			(*out_unmanaged_changed)++;
		else
			_nm_settings_plugin_emit_signal_unmanaged_specs_changed (NM_SETTINGS_PLUGIN (etcnet_device->plugin));
	}
}

static gboolean
is_wifi_device (const char *name)
{
	int ifindex;

	g_return_val_if_fail (name != NULL, FALSE);

	ifindex = nm_platform_link_get_ifindex (NM_PLATFORM_GET, name);
	if (ifindex == 0)
		return FALSE;

	return nm_platform_link_get_type (NM_PLATFORM_GET, ifindex) == NM_LINK_TYPE_WIFI;
}

static NMDeviceType
get_dev_type(const char *iface, struct udev_device *device)
{
	const char *sysfs_path;
	const char *type_str;

	g_return_val_if_fail (iface != NULL, NM_DEVICE_TYPE_UNKNOWN);
	g_return_val_if_fail (device != NULL, NM_DEVICE_TYPE_UNKNOWN);

	/* Check device type */
	type_str = udev_device_get_sysattr_value (device, "type");
	if (!nm_streq0 (type_str, "1"))
		return NM_DEVICE_TYPE_UNKNOWN;

	/* Check if device is bridge */
	sysfs_path = udev_device_get_syspath (device);

	if (sysfs_path && g_str_has_prefix (sysfs_path, "/sys/devices/virtual")) {
		char bridge_path[255];
		struct stat s;

		g_snprintf (bridge_path, sizeof (bridge_path), "%s/bridge", sysfs_path);
		if (stat (bridge_path, &s) == 0 && (s.st_mode & S_IFDIR))
			return NM_DEVICE_TYPE_BRIDGE;
	}

	return is_wifi_device (iface) ? NM_DEVICE_TYPE_WIFI : NM_DEVICE_TYPE_ETHERNET;
}

static void
read_device_config(SettingsPluginEtcnet *plugin, struct udev_device *device, guint *out_unmanaged_changed)
{
	NMEtcnetDevice *etcnet_device;
	const char *iface;
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);

	g_return_if_fail (plugin != NULL);
	g_return_if_fail (device != NULL);

	iface = udev_device_get_sysname (device);
	g_return_if_fail (iface != NULL);

	etcnet_device = g_hash_table_lookup(priv->devices, iface);
	if(!etcnet_device) {
		NMDeviceType dev_type = get_dev_type(iface, device);

		etcnet_device = nm_etcnet_device_new(plugin, priv->ifaces_path, iface, dev_type);
		if (etcnet_device == NULL)
			return;

		g_hash_table_insert(priv->devices, g_strdup(iface), etcnet_device);
	}

	update_device_config (etcnet_device, out_unmanaged_changed);
}

static void
read_connections (SettingsPluginEtcnet *plugin)
{
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);
	struct udev_enumerate *enumerate;
	struct udev_list_entry *keys;
	guint unmanaged_changed = 0;

	enumerate = nm_udev_client_enumerate_new (priv->udev_client);
	g_return_if_fail (enumerate != NULL);

	udev_enumerate_scan_devices (enumerate);
	keys = udev_enumerate_get_list_entry (enumerate);
	for (; keys; keys = udev_list_entry_get_next (keys)) {
		struct udev_device *udevice;

		udevice = udev_device_new_from_syspath (udev_enumerate_get_udev (enumerate),
		                                        udev_list_entry_get_name (keys));
		if (udevice) {
			read_device_config (plugin, udevice, &unmanaged_changed);
			udev_device_unref (udevice);
		}
	}

	if (unmanaged_changed)
		_nm_settings_plugin_emit_signal_unmanaged_specs_changed (NM_SETTINGS_PLUGIN (plugin));

	udev_enumerate_unref (enumerate);
}

static void
get_connections_cb (gpointer key, gpointer val, gpointer user_data)
{
	GSList **list = (GSList **) user_data;

	*list = g_slist_prepend (*list, val);
}

static void
get_devices_cb (gpointer key, gpointer val, gpointer user_data)
{
	NMEtcnetDevice *etcnet_device = (NMEtcnetDevice *) val;

	if (etcnet_device->connections)
		g_hash_table_foreach (etcnet_device->connections, get_connections_cb, user_data);
}

static GSList *
get_connections (NMSettingsPlugin *config)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (config);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);
	GSList *list = NULL;

	if(!g_hash_table_size(priv->devices)) {
		read_connections (plugin);
	}

	g_hash_table_foreach (priv->devices, get_devices_cb, &list);

	return list;
}

static void
reload_connections (NMSettingsPlugin *config)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (config);

	read_connections (plugin);
}

static void
handle_uevent_cb (NMUdevClient *client,
		struct udev_device *device,
		gpointer user_data)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (user_data);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);
	const char *subsys;
	const char *action;
	NMEtcnetDevice *etcnet_device;

	action = udev_device_get_action (device);
	g_return_if_fail (action != NULL);

	/* A bit paranoid */
	subsys = udev_device_get_subsystem (device);
	g_return_if_fail (nm_streq0 (subsys, "net"));

	if (!strcmp ("add", action)) {
		read_device_config (plugin, device, NULL);
	} else if (!strcmp ("remove", action)) {
		const char *iface;

		iface = udev_device_get_sysname (device);
		g_return_if_fail (iface != NULL);

		etcnet_device = (NMEtcnetDevice *)g_hash_table_lookup(priv->devices, iface);
		if(etcnet_device)
			g_hash_table_remove (priv->devices, iface);
	}
}

static void
initialize (NMSettingsPlugin *config)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (config);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);
	const char *subsys[2] = { "net", NULL };

	priv->ifaces_path = g_strdup(ETCNET_IFACES_PATH);
	if (!priv->ifaces_path)
		PLUGIN_WARNING("could not determine network ifaces path.");

	priv->udev_client = nm_udev_client_new (subsys, handle_uevent_cb, plugin);

	priv->need_monitor_files = nm_config_get_monitor_connection_files (nm_config_get ());

	priv->current_profile = get_current_profile();
	PLUGIN_DEBUG ("Current network profile: %s",
			priv->current_profile && priv->current_profile[0] != '#' ?
			priv->current_profile : "(none)");

	/* All pivate data should be automatically zero'ed when the GObject is created,
	 * so nothing to do in case if need_monitor_files is FALSE */
	if (priv->need_monitor_files) {
		start_monitor (&priv->ifaces_monitor, priv->ifaces_path, G_CALLBACK(ifaces_dir_changed_cb), plugin);
	}
}

static void
settings_plugin_etcnet_init (SettingsPluginEtcnet *plugin)
{
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);

	priv->devices = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, nm_etcnet_device_free);
}

static void
dispose (GObject *object)
{
	SettingsPluginEtcnet *plugin = SETTINGS_PLUGIN_ETCNET (object);
	SettingsPluginEtcnetPrivate *priv = SETTINGS_PLUGIN_ETCNET_GET_PRIVATE (plugin);

	priv->udev_client = nm_udev_client_unref (priv->udev_client);
	g_hash_table_destroy (priv->devices);

	g_free(priv->ifaces_path);
	if (priv->need_monitor_files) {
		stop_monitor (&priv->ifaces_monitor);
	}

	g_free (priv->current_profile);

	G_OBJECT_CLASS (settings_plugin_etcnet_parent_class)->dispose (object);
}

static void
settings_plugin_etcnet_class_init (SettingsPluginEtcnetClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	NMSettingsPluginClass *plugin_class = NM_SETTINGS_PLUGIN_CLASS (klass);

	object_class->dispose = dispose;

	plugin_class->get_connections = get_connections;
	plugin_class->reload_connections = reload_connections;
	plugin_class->get_unmanaged_specs = get_unmanaged_specs;
	plugin_class->initialize = initialize;
}

G_MODULE_EXPORT NMSettingsPlugin *
nm_settings_plugin_factory (void)
{
	return NM_SETTINGS_PLUGIN (g_object_ref (settings_plugin_etcnet_get ()));
}
